@@ -5,6 +5,7 @@ import numpy as np
 import pandas as pd
 from gensim.models.doc2vec import Doc2Vec, TaggedDocument
 from gensim.utils import simple_preprocess
+from gensim.parsing.preprocessing import strip_tags
 import umap
 import hdbscan
 from operator import itemgetter
@@ -12,6 +13,7 @@ from sklearn.metrics.pairwise import cosine_similarity
 from wordcloud import WordCloud
 import matplotlib.pyplot as plt
 from joblib import dump, load
+from sklearn.cluster import dbscan
 
 
 class Top2Vec:
@@ -90,53 +92,113 @@ class Top2Vec:
         else:
             raise ValueError("workers needs to be an int")
 
-        self.documents = documents
+        self.documents = list(documents)
 
         # preprocess documents for training - tokenize and remove too long/short words
-        train_corpus = [TaggedDocument(simple_preprocess(doc), [i]) for i, doc in enumerate(documents)]
+        train_corpus = [TaggedDocument(simple_preprocess(strip_tags(doc), deacc=True), [i])
+                        for i, doc in enumerate(documents)]
 
         # create documents and word embeddings with doc2vec
         if workers is None:
-            self.model = Doc2Vec(documents=train_corpus, vector_size=300, min_count=50, window=15,
-                                 sample=10e-5, negative=negative, hs=hs, epochs=epochs, dm=0,
+            self.model = Doc2Vec(documents=train_corpus,
+                                 vector_size=300,
+                                 min_count=50, window=15,
+                                 sample=1e-5,
+                                 negative=negative,
+                                 hs=hs,
+                                 epochs=epochs,
+                                 dm=0,
                                  dbow_words=1)
         else:
-            self.model = Doc2Vec(documents=train_corpus, vector_size=300, min_count=50, window=15,
-                                 sample=10e-5, negative=negative, hs=hs, workers=workers, epochs=epochs, dm=0,
+            self.model = Doc2Vec(documents=train_corpus,
+                                 vector_size=300,
+                                 min_count=50,
+                                 window=15,
+                                 sample=1e-5,
+                                 negative=negative,
+                                 hs=hs,
+                                 workers=workers,
+                                 epochs=epochs,
+                                 dm=0,
                                  dbow_words=1)
 
         # create 5D embeddings of documents
-        docvecs = np.vstack([self.model.docvecs[i] for i in range(self.model.docvecs.count)])
-        umap_model = umap.UMAP(n_neighbors=15, n_components=5, metric='cosine').fit(docvecs)
+        umap_model = umap.UMAP(n_neighbors=15,
+                               n_components=5,
+                               metric='cosine').fit(self.model.docvecs.vectors_docs)
 
         # find dense areas of document vectors
-        cluster = hdbscan.HDBSCAN(min_cluster_size=15, metric='euclidean', cluster_selection_method='eom').fit(
-            umap_model.embedding_)
+        cluster = hdbscan.HDBSCAN(min_cluster_size=15,
+                                  metric='euclidean',
+                                  cluster_selection_method='eom').fit(umap_model.embedding_)
 
-        cluster_labels = pd.Series(cluster.labels_)
+        # calculate topic vectors from dense areas of documents
+        self._create_topic_vectors(cluster.labels_)
 
-        # generate topic vectors from dense areas of documents
-        self.topic_vectors = []
+        # deduplicate topics
+        self._deduplicate_topics()
+
+        # calculate topic sizes and index nearest topic for each document
+        self._calculate_topic_sizes()
+
+        # find topic words and scores
+        self._find_topic_words_scores()
+
+    def _create_topic_vectors(self, cluster_labels):
+
+        unique_labels = set(cluster_labels)
+        if -1 in unique_labels:
+            unique_labels.remove(-1)
+        self.topic_vectors = np.vstack([self.model.docvecs.vectors_docs[np.where(cluster_labels == label)[0]]
+                                       .mean(axis=0) for label in unique_labels])
+
+    def _deduplicate_topics(self):
+        core_samples, labels = dbscan(X=self.topic_vectors,
+                                      eps=0.1,
+                                      min_samples=2,
+                                      metric="cosine")
+
+        duplicate_clusters = set(labels)
+
+        if len(duplicate_clusters) > 1 or -1 not in duplicate_clusters:
+
+            # unique topics
+            unique_topics = self.topic_vectors[np.where(labels == -1)[0]]
+
+            if -1 in duplicate_clusters:
+                duplicate_clusters.remove(-1)
+
+            # merge duplicate topics
+            for unique_label in duplicate_clusters:
+                unique_topics = np.vstack(
+                    [unique_topics, self.topic_vectors[np.where(labels == unique_label)[0]]
+                        .mean(axis=0)])
+
+            self.topic_vectors = unique_topics
+
+    def _calculate_topic_sizes(self):
+        # calculate topic size
+        doc_top_sim = cosine_similarity(self.model.docvecs.vectors_docs, self.topic_vectors)
+        self.topic_sizes = pd.Series(np.argmax(doc_top_sim, axis=1)).value_counts()
+
+        # re-order topic vectors by size
+        self.topic_vectors = self.topic_vectors[self.topic_sizes.index]
+        doc_top_sim = doc_top_sim[:, self.topic_sizes.index]
+        self.topic_sizes.reset_index(drop=True, inplace=True)
+
+        # find nearest topic for each document and distance to topic
+        self.doc_dist = np.max(doc_top_sim, axis=1)
+        self.doc_top = np.argmax(doc_top_sim, axis=1)
+
+    def _find_topic_words_scores(self):
         self.topic_words = []
         self.topic_word_scores = []
+        np.apply_along_axis(self._generate_topic_words_scores, axis=1, arr=self.topic_vectors)
 
-        unique_labels = list(set(cluster_labels))
-        unique_labels.remove(-1)
-
-        for label in unique_labels:
-
-            # find centroid of dense document cluster
-            topic_vector = [0] * 300
-            cluster_vec_indices = cluster_labels[cluster_labels == label].index.tolist()
-            for vec_index in cluster_vec_indices:
-                topic_vector = topic_vector + self.model.docvecs[vec_index]
-            topic_vector = topic_vector / len(cluster_vec_indices)
-            self.topic_vectors.append(topic_vector)
-
-            # find closest word vectors to topic vector
-            sim_words = self.model.most_similar(positive=[topic_vector], topn=50)
-            self.topic_words.append([word[0] for word in sim_words])
-            self.topic_word_scores.append([round(word[1], 4) for word in sim_words])
+    def _generate_topic_words_scores(self, topic_vector):
+        sim_words = self.model.wv.most_similar(positive=[topic_vector], topn=50)
+        self.topic_words.append([word[0] for word in sim_words])
+        self.topic_word_scores.append([round(word[1], 4) for word in sim_words])
 
     def save(self, file):
         """
@@ -171,7 +233,7 @@ class Top2Vec:
         self._less_than_zero(num_docs, "num_docs")
         document_count = len(self.documents)
         if num_docs > document_count:
-            raise ValueError(f"num_docs cannot exceed the number of topics: {document_count}")
+            raise ValueError(f"num_docs cannot exceed the number of documents: {document_count}")
 
     def _validate_num_topics(self, num_topics):
         self._less_than_zero(num_topics, "num_topics")
@@ -181,15 +243,20 @@ class Top2Vec:
 
     def _validate_topic_num(self, topic_num):
         self._less_than_zero(topic_num, "topic_num")
-        topic_count = len(self.topic_vectors)-1
+        topic_count = len(self.topic_vectors) - 1
         if topic_num > topic_count:
             raise ValueError(f"Invalid topic number: valid topics numbers are 0 to {topic_count}")
 
+    def _validate_topic_search(self, topic_num, doc_num):
+        if doc_num > self.topic_sizes[topic_num]:
+            raise ValueError(f"Invalid number of documents: topic {topic_num}"
+                             f" only has {self.topic_sizes[topic_num]} documents")
+
     def _validate_doc_num(self, doc_num):
         self._less_than_zero(doc_num, "doc_num")
-        document_count = len(self.documents)-1
+        document_count = len(self.documents) - 1
         if doc_num > document_count:
-            raise ValueError(f"Invalid topic number: valid topic numbers are 0 to {document_count}")
+            raise ValueError(f"Invalid document number: valid document numbers are 0 to {document_count}")
 
     def _validate_keywords(self, keywords, keywords_neg):
 
@@ -199,7 +266,7 @@ class Top2Vec:
         if not isinstance(keywords_neg, list):
             raise ValueError("keywords_neg must be a list of strings")
 
-        for word in keywords+keywords_neg:
+        for word in keywords + keywords_neg:
             if word not in self.model.wv.vocab:
                 raise ValueError(f"'{word}' has not been learned by the model so it cannot be searched")
 
@@ -215,9 +282,25 @@ class Top2Vec:
         """
         return len(self.topic_vectors)
 
+    def get_topic_sizes(self):
+        """
+        Get topic sizes.
+
+        The number of documents most similar to each topic. Topics are
+        in increasing order of size.
+
+        Returns
+        -------
+        topic_sizes: list
+            The number of documents most similar to the topic.
+        topic_nums:
+            The unique index of every topic will be returned.
+        """
+        return self.topic_sizes.values.tolist(), self.topic_sizes.index.tolist()
+
     def get_topics(self, num_topics):
         """
-        Get number of specified topics.
+        Get number of specified topics, ordered by decreasing size.
 
         Each topic will consist of the top 50 semantically similar words
         to the topic. These are the 50 words closest to topic vector
@@ -286,10 +369,13 @@ class Top2Vec:
 
         self._validate_num_docs(num_docs)
         self._validate_topic_num(topic_num)
+        self._validate_topic_search(topic_num, num_docs)
+
+        topic_document_indexes = np.where(self.doc_top == topic_num)[0]
+        topic_document_indexes_ordered = np.flip(np.argsort(self.doc_dist[topic_document_indexes]))
+        doc_nums = topic_document_indexes[topic_document_indexes_ordered][0:num_docs]
+        doc_scores = self.doc_dist[doc_nums]
 
-        sim_docs = self.model.docvecs.most_similar(positive=[self.topic_vectors[topic_num]], topn=num_docs)
-        doc_nums = [doc[0] for doc in sim_docs]
-        doc_scores = [round(doc[1], 4) for doc in sim_docs]
         documents = list(itemgetter(*doc_nums)(self.documents))
 
         return documents, doc_scores, doc_nums
@@ -335,7 +421,9 @@ class Top2Vec:
 
         word_vecs = [self.model[word] for word in keywords]
         neg_word_vecs = [self.model[word] for word in keywords_neg]
-        sim_docs = self.model.docvecs.most_similar(positive=word_vecs, negative=neg_word_vecs, topn=num_docs)
+        sim_docs = self.model.docvecs.most_similar(positive=word_vecs,
+                                                   negative=neg_word_vecs,
+                                                   topn=num_docs)
         doc_nums = [doc[0] for doc in sim_docs]
         doc_scores = [round(doc[1], 4) for doc in sim_docs]
         documents = list(itemgetter(*doc_nums)(self.documents))
@@ -380,9 +468,11 @@ class Top2Vec:
 
         word_vecs = [self.model[word] for word in keywords]
         neg_word_vecs = [self.model[word] for word in keywords_neg]
-        sim_words = self.model.most_similar(positive=word_vecs, negative=neg_word_vecs, topn=num_words)
+        sim_words = self.model.wv.most_similar(positive=word_vecs,
+                                               negative=neg_word_vecs,
+                                               topn=num_words)
         words = [word[0] for word in sim_words]
-        word_scores = [round(word[1],4) for word in sim_words]
+        word_scores = [round(word[1], 4) for word in sim_words]
 
         return words, word_scores
 
@@ -439,7 +529,7 @@ class Top2Vec:
         word_vecs = [self.model[word] for word in keywords]
         neg_word_vecs = [self.model[word] for word in keywords_neg]
 
-        combined_vector = [0] * 300
+        combined_vector = np.zeros(300, dtype=np.float64)
 
         for word_vec in word_vecs:
             combined_vector += word_vec
@@ -450,8 +540,7 @@ class Top2Vec:
         combined_vector /= (len(word_vecs) + len(neg_word_vecs))
 
         topic_ranks = [topic[0] for topic in cosine_similarity(self.topic_vectors, combined_vector.reshape(1, -1))]
-        topic_nums = list(np.argsort(topic_ranks)[-num_topics:])
-        topic_nums.reverse()
+        topic_nums = list(np.flip(np.argsort(topic_ranks)[-num_topics:]))
 
         topic_words = [self.topic_words[topic] for topic in topic_nums]
         word_scores = [self.topic_word_scores[topic] for topic in topic_nums]
@@ -491,14 +580,15 @@ class Top2Vec:
         self._validate_doc_num(doc_num)
 
         message_vec = self.model.docvecs[doc_num]
-        sim_docs = self.model.docvecs.most_similar(positive=[message_vec], topn=num_docs)
+        sim_docs = self.model.docvecs.most_similar(positive=[message_vec],
+                                                   topn=num_docs)
         doc_nums = [doc[0] for doc in sim_docs]
         doc_scores = [round(doc[1], 4) for doc in sim_docs]
         documents = list(itemgetter(*doc_nums)(self.documents))
 
         return documents, doc_scores, doc_nums
 
-    def generate_topic_wordcloud(self, topic_num, background_color="white"):
+    def generate_topic_wordcloud(self, topic_num, background_color="black"):
         """
         Create a word cloud for a topic.
 
@@ -525,9 +615,11 @@ class Top2Vec:
         self._validate_topic_num(topic_num)
 
         word_score_dict = dict(zip(self.topic_words[topic_num], self.topic_word_scores[topic_num]))
-        plt.figure(figsize=(16, 4), dpi=200)
+        plt.figure(figsize=(16, 4),
+                   dpi=200)
         plt.axis("off")
         plt.imshow(
-            WordCloud(width=1600, height=400, background_color=background_color).generate_from_frequencies(word_score_dict));
+            WordCloud(width=1600,
+                      height=400,
+                      background_color=background_color).generate_from_frequencies(word_score_dict))
         plt.title("Topic " + str(topic_num), loc='left', fontsize=25, pad=20)
-
